//----------------------------------------------------------------------------
// Copyright (c) 2007-2008 Fabrice HARROUET (ENIB)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.
// The author makes no representations about the suitability of this
// software for any purpose.
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------------

#include "netUtils.h"

#include <pthread.h>

#include <openssl/ssl.h>
#include <openssl/err.h>

#include <iostream>
#include <vector>
using namespace std;

//----------------------------------------------------------------------------

pthread_mutex_t * ssl_locks;

void
ssl_locking_cb(int mode,
               int n,
               const char * file,
               int line)
{
(void)file;
(void)line;
if(mode&CRYPTO_LOCK)
  {
  pthread_mutex_lock(ssl_locks+n);
  }
else
  {
  pthread_mutex_unlock(ssl_locks+n);
  }
}

unsigned long
ssl_id_cb(void)
{
return (unsigned long)pthread_self();
}

MAKE_READLINE_FUNCTION(sslReadLine,SSL_read,SSL *)

MAKE_WRITEFULLY_FUNCTION(sslWriteFully,SSL_write,SSL *)

//----------------------------------------------------------------------------

void
waitChildren(int signum)
{
int r;
(void)signum;
do
  {
  RESTART_SYSCALL(r,waitpid(-1,(int *)0,WNOHANG));
  if(r!=-1)
    {
    fprintf(stderr,"Child %d done\n",r);
    }
  } while(r!=-1);
}

//----------------------------------------------------------------------------

#define BUFFER_SIZE 0x400

class BounceClient;

class BounceServer
{
public:

  BounceServer(unsigned long realIp,
               int realPort,
               int listenPort);

  virtual
  ~BounceServer(void);

  virtual
  int
  getListenFd(void);

  virtual
  void
  process(void)=0;

  static
  void
  startThread(BounceClient * client);

protected:

  static
  void *
  _thread(void * data);

  friend class BounceClient;

  unsigned long _realIp;
  int _realPort;
  int _listenPort;
  int _listenFd;
};

class TcpBounceClient;

class TcpBounceServer : public BounceServer
{
public:

  TcpBounceServer(unsigned long realIp,
                  int realPort,
                  int listenPort);

  virtual
  ~TcpBounceServer(void);

  virtual
  void
  process(void);

protected:

  friend class TcpBounceClient;

};

class SslBounceClient;

class SslBounceServer : public TcpBounceServer
{
public:

  SslBounceServer(SSL_CTX * ctx,
                  unsigned long realIp,
                  int realPort,
                  int listenPort);

  virtual
  ~SslBounceServer(void);

  virtual
  void
  process(void);

protected:

  friend class SslBounceClient;

  SSL_CTX * _ctx;
};

class BounceClient
{
public:

  BounceClient(BounceServer * server);

  virtual
  ~BounceClient(void);

  virtual
  void
  process(void)=0;

protected:

  virtual
  unsigned long
  _realIp(void);

  virtual
  int
  _realPort(void);

  virtual
  int
  _listenPort(void);

  virtual
  int
  _openDumpFile(const char * prefix);

  BounceServer * _server;
  unsigned long _clientIp;
  int _clientPort;
};

class TcpBounceClient : public BounceClient
{
public:

  TcpBounceClient(TcpBounceServer * server);

  virtual
  ~TcpBounceClient(void);

  virtual
  void
  process(void);

protected:

  int _clientFd;
  int _serverFd;
};

class SslBounceClient : public TcpBounceClient
{
public:

  SslBounceClient(SslBounceServer * server);

  virtual
  ~SslBounceClient(void);

  virtual
  void
  process(void);

protected:

  SSL * _clientSsl;
  SSL * _serverSsl;
};

//----------------------------------------------------------------------------

BounceServer::BounceServer(unsigned long realIp,
                           int realPort,
                           int listenPort)
: _realIp(realIp),
  _realPort(realPort),
  _listenPort(listenPort),
  _listenFd(-1)
{
}

BounceServer::~BounceServer(void)
{
if(_listenFd!=-1)
  {
  int r;
  RESTART_SYSCALL(r,close(_listenFd));
  }
}

int
BounceServer::getListenFd(void)
{
return _listenFd;
}

// void
// BounceServer::process(void)=0

void
BounceServer::startThread(BounceClient * client) // static
{
pthread_t th;
int r=pthread_create(&th,(pthread_attr_t *)0,&BounceServer::_thread,client);
if(r)
  {
  cerr << "pthread_create: " << strerror(r) << endl;
  delete client;
  }
}

void *
BounceServer::_thread(void * data) // static
{
BounceClient * client=(BounceClient *)data;
pthread_detach(pthread_self());
client->process();
delete client;
return (void *)0;
}

TcpBounceServer::TcpBounceServer(unsigned long realIp,
                                 int realPort,
                                 int listenPort)
: BounceServer(realIp,realPort,listenPort)
{
_listenFd=listenTcp(_listenPort);
if(_listenFd==-1)
  {
  cerr << "listenTcp on port " << _listenPort << ": "
       << strerror(errno) << endl;
  }
}

TcpBounceServer::~TcpBounceServer(void)
{
}

void
TcpBounceServer::process(void)
{
BounceServer::startThread(new TcpBounceClient(this));
}


SslBounceServer::SslBounceServer(SSL_CTX * ctx,
                                 unsigned long realIp,
                                 int realPort,
                                 int listenPort)
: TcpBounceServer(realIp,realPort,listenPort),
  _ctx(ctx)
{
}

SslBounceServer::~SslBounceServer(void)
{
}

void
SslBounceServer::process(void)
{
BounceServer::startThread(new SslBounceClient(this));
}

BounceClient::BounceClient(BounceServer * server)
: _server(server),
  _clientIp(0),
  _clientPort(0)
{
}

BounceClient::~BounceClient(void)
{
}

// void
// BounceClient::process(void)=0

unsigned long
BounceClient::_realIp(void)
{
return _server->_realIp;
}

int
BounceClient::_realPort(void)
{
return _server->_realPort;
}

int
BounceClient::_listenPort(void)
{
return _server->_listenPort;
}

int
BounceClient::_openDumpFile(const char * prefix)
{
char fileName[0x100];
char server[0x10];
char client[0x10];
struct timeval tv;
gettimeofday(&tv,(struct timezone *)0);
struct tm * now=localtime(&tv.tv_sec);
writeIpAddress(server,0x10,_server->_realIp);
writeIpAddress(client,0x10,_clientIp);
sprintf(fileName,"%s___S_%s_%d___C_%s_%d___"
                 "D_%.4d_%.2d_%.2d___T_%.2d_%.2d_%.2d.%.3ld",
                 prefix,server,_realPort(),client,_clientPort,
                 now->tm_year+1900,now->tm_mon+1,now->tm_mday,
                 now->tm_hour,now->tm_min,now->tm_sec,tv.tv_usec/1000);
int fd;
RESTART_SYSCALL(fd,open(fileName,O_WRONLY|O_TRUNC|O_CREAT,0644));
if(fd==-1)
  {
  cerr << "Cannot open " << fileName << endl;
  }
else
  {
  cerr << fileName << endl;
  }
return fd;
}

TcpBounceClient::TcpBounceClient(TcpBounceServer * server)
: BounceClient(server),
  _clientFd(acceptTcp(_server->getListenFd(),&_clientIp,&_clientPort)),
  _serverFd(-1)
{
if(_clientFd==-1)
  {
  cerr << "acceptTcp on port " << _listenPort() << ": "
       << strerror(errno) << endl;
  }
}

TcpBounceClient::~TcpBounceClient(void)
{
int r;
if(_clientFd!=-1)
  {
  RESTART_SYSCALL(r,close(_clientFd));
  }
if(_serverFd!=-1)
  {
  RESTART_SYSCALL(r,close(_serverFd));
  }
}

void
TcpBounceClient::process(void)
{
int r;
if(_clientFd==-1)
  {
  return;
  }
_serverFd=connectTcp(_realIp(),_realPort());
if(_serverFd==-1)
  {
  char server[0x10];
  writeIpAddress(server,0x10,_realIp());
  cerr << "connectTcp to " << server << " on port " << _realPort() << ": "
       << strerror(errno) << endl;
  return;
  }
int dumpFd=_openDumpFile("TCP");
if(dumpFd==-1)
  {
  return;
  }
char buffer[BUFFER_SIZE];
for(;;)
  {
  int maxFd=-1;
  fd_set rdSet;
  FD_ZERO(&rdSet);
  FD_SET_MAX(_clientFd,&rdSet,maxFd);
  FD_SET_MAX(_serverFd,&rdSet,maxFd);
  RESTART_SYSCALL(r,select(maxFd+1,&rdSet,(fd_set *)0,(fd_set *)0,
                           (struct timeval *)0));
  if(r==-1)
    {
    char server[0x10];
    writeIpAddress(server,0x10,_realIp());
    cerr << "select for " << server << " on port " << _realPort() << ": "
         << strerror(errno) << endl;
    break;
    }
  if(FD_ISSET(_clientFd,&rdSet))
    {
    RESTART_SYSCALL(r,read(_clientFd,buffer,BUFFER_SIZE));
    if(r>0)
      {
      writeFully(dumpFd,buffer,r);
      writeFully(_serverFd,buffer,r);
      }
    else
      {
      break;
      }
    }
  if(FD_ISSET(_serverFd,&rdSet))
    {
    RESTART_SYSCALL(r,read(_serverFd,buffer,BUFFER_SIZE));
    if(r>0)
      {
      writeFully(dumpFd,buffer,r);
      writeFully(_clientFd,buffer,r);
      }
    else
      {
      break;
      }
    }
  }
RESTART_SYSCALL(r,close(dumpFd));
}

SslBounceClient::SslBounceClient(SslBounceServer * server)
: TcpBounceClient(server),
  _clientSsl(SSL_new(server->_ctx)),
  _serverSsl(SSL_new(server->_ctx))
{
}

SslBounceClient::~SslBounceClient(void)
{
SSL_free(_clientSsl);
SSL_free(_serverSsl);
}

void
SslBounceClient::process(void)
{
int r;
if(_clientFd==-1)
  {
  cerr << "acceptTcp on port " << _listenPort() << ": "
       << strerror(errno) << endl;
  return;
  }
SSL_set_fd(_clientSsl,_clientFd);
SSL_set_mode(_clientSsl,SSL_MODE_AUTO_RETRY);
r=SSL_accept(_clientSsl);
if(r!=1)
  {
  cerr << "SSL_accept on port " << _listenPort() << ": "
       << ERR_error_string(ERR_get_error(),NULL) << endl;
  return;
  }
_serverFd=connectTcp(_realIp(),_realPort());
if(_serverFd==-1)
  {
  char server[0x10];
  writeIpAddress(server,0x10,_realIp());
  cerr << "connectTcp to " << server << " on port " << _realPort() << ": "
       << strerror(errno) << endl;
  return;
  }
SSL_set_fd(_serverSsl,_serverFd);
SSL_set_mode(_serverSsl,SSL_MODE_AUTO_RETRY);
r=SSL_connect(_serverSsl);
if(r!=1)
  {
  char server[0x10];
  writeIpAddress(server,0x10,_realIp());
  cerr << "SSL_connect to " << server << " on port " << _realPort() << ": "
       << ERR_error_string(ERR_get_error(),NULL) << endl;
  return;
  }
int dumpFd=_openDumpFile("SSL");
if(dumpFd==-1)
  {
  return;
  }
char buffer[BUFFER_SIZE];
for(;;)
  {
  int maxFd=-1;
  fd_set rdSet;
  struct timeval tv={0,0};
  struct timeval * ptv=(struct timeval *)0;
  FD_ZERO(&rdSet);
  if(SSL_pending(_clientSsl)) ptv=&tv;
  else FD_SET_MAX(_clientFd,&rdSet,maxFd);
  if(SSL_pending(_serverSsl)) ptv=&tv;
  else FD_SET_MAX(_serverFd,&rdSet,maxFd);
  RESTART_SYSCALL(r,select(maxFd+1,&rdSet,(fd_set *)0,(fd_set *)0,ptv));
  if(r==-1)
    {
    char server[0x10];
    writeIpAddress(server,0x10,_realIp());
    cerr << "select for " << server << " on port " << _realPort() << ": "
         << strerror(errno) << endl;
    break;
    }
  if(SSL_pending(_clientSsl)||FD_ISSET(_clientFd,&rdSet))
    {
    RESTART_SYSCALL(r,SSL_read(_clientSsl,buffer,BUFFER_SIZE));
    if(r>0)
      {
      writeFully(dumpFd,buffer,r);
      sslWriteFully(_serverSsl,buffer,r);
      }
    else
      {
      break;
      }
    }
  if(SSL_pending(_serverSsl)||FD_ISSET(_serverFd,&rdSet))
    {
    RESTART_SYSCALL(r,SSL_read(_serverSsl,buffer,BUFFER_SIZE));
    if(r>0)
      {
      writeFully(dumpFd,buffer,r);
      sslWriteFully(_clientSsl,buffer,r);
      }
    else
      {
      break;
      }
    }
  }
RESTART_SYSCALL(r,close(dumpFd));
}

//----------------------------------------------------------------------------

void
displayUsage(const string & cmd)
{
string spc=string(cmd.size(),' ');
cerr <<
"usage: " << cmd << " [ -tcp @IP:serverPort:localPort] ... \\\n"
"       " << spc << " [ -ssl @IP:serverPort:localPort] ...\n"
"\n"
"    -tcp @IP:serverPort:localPort                        (0 or many)\n"
"        redirect and dump TCP traffic on localPort to @IP:serverPort\n"
"\n"
"    -ssl @IP:serverPort:localPort                        (0 or many)\n"
"        redirect and dump SSL traffic on localPort to @IP:serverPort\n"
<< endl;
}

bool
parseArgs(vector<BounceServer *> & servers,
          SSL_CTX * ctx,
          int argc,
          char ** argv)
{
bool showUsage=false;
bool success=true;
for(int i=1;i<argc;i++)
  {
  if(!strcmp(argv[i],"-tcp")||!strcmp(argv[i],"-ssl"))
    {
    if(++i>=argc)
      {
      showUsage=true;
      }
    else
      {
      bool ok=false;
      string server;
      int serverPort,localPort;
      char * c1=strchr(argv[i],':');
      if(c1)
        {
        server=string(argv[i],c1-argv[i]);
        char * c2=strchr(c1+1,':');
        if(c2&&(sscanf(c1+1,"%d",&serverPort)==1))
          {
          if(sscanf(c2+1,"%d",&localPort)==1)
            {
            ok=true;
            }
          }
        }
      if(ok)
        {
        unsigned long addr=getIpAddress(server.c_str());
        if(addr)
          {
          if(!strcmp(argv[i-1],"-tcp"))
            {
            servers.push_back(new TcpBounceServer(addr,serverPort,
                                                  localPort));
            }
          else
            {
            servers.push_back(new SslBounceServer(ctx,addr,serverPort,
                                                  localPort));
            }
          if(servers.back()->getListenFd()==-1)
            {
            delete servers.back();
            servers.pop_back();
            success=false;
            }
          }
        else
          {
          cerr << "Unknown host `" << server << "'" << endl;
          success=false;
          }
        }
      else
        {
        showUsage=true;
        }
      }
    }
  else
    {
    showUsage=true;
    }
  }
if(showUsage||servers.empty())
  {
  displayUsage(argv[0]);
  success=false;
  }
return success;
}

int
main(int argc,
     char ** argv)
{
cerr << endl << "Bounce ..." << endl;
cerr << "(dump man-in-the-middle traffic)" << endl << endl;

size_t i;
int r;
struct sigaction act;
memset(&act,0,sizeof(struct sigaction));
act.sa_handler=&waitChildren;
RESTART_SYSCALL(r,sigaction(SIGCHLD,&act,(struct sigaction *)0));
act.sa_handler=SIG_IGN;
RESTART_SYSCALL(r,sigaction(SIGPIPE,&act,(struct sigaction *)0));
if(r==-1)
  {
  perror("sigaction");
  exit(EXIT_FAILURE);
  }
SSL_library_init();
SSL_load_error_strings();
OpenSSL_add_all_algorithms();
SSL_CTX * ctx=SSL_CTX_new(SSLv23_method());
ssl_locks=(pthread_mutex_t *)malloc(CRYPTO_num_locks()*
                                    sizeof(pthread_mutex_t));
for(i=CRYPTO_num_locks();i--;)
  {
  pthread_mutex_init(ssl_locks+i,(pthread_mutexattr_t *)0);
  }
CRYPTO_set_locking_callback(&ssl_locking_cb);
CRYPTO_set_id_callback(&ssl_id_cb);
SSL_CTX_use_certificate_file(ctx,"cert.pem",SSL_FILETYPE_PEM);
SSL_CTX_use_PrivateKey_file(ctx,"key.pem",SSL_FILETYPE_PEM);
if(!SSL_CTX_check_private_key(ctx))
  {
  cerr << "SSL_CTX_check_private_key: "
       << ERR_error_string(ERR_get_error(),NULL) << endl;
  exit(EXIT_FAILURE);
  }
vector<BounceServer *> servers;
if(!parseArgs(servers,ctx,argc,argv))
  {
  exit(EXIT_FAILURE);
  }
for(;;)
  {
  int maxFd=-1;
  fd_set rdSet;
  FD_ZERO(&rdSet);
  for(i=servers.size();i--;)
    {
    FD_SET_MAX(servers[i]->getListenFd(),&rdSet,maxFd);
    }
  RESTART_SYSCALL(r,select(maxFd+1,&rdSet,(fd_set *)0,(fd_set *)0,
                           (struct timeval *)0));
  if(r==-1)
    {
    cerr << "select: " << strerror(errno) << endl;
    break;
    }
  for(i=servers.size();i--;)
    {
    if(FD_ISSET(servers[i]->getListenFd(),&rdSet))
      {
      servers[i]->process();
      }
    }
  }
return EXIT_SUCCESS;
}

//----------------------------------------------------------------------------
