//----------------------------------------------------------------------------
// Copyright (c) 2007-2008 Fabrice HARROUET (ENIB)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.
// The author makes no representations about the suitability of this
// software for any purpose.
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------------

#include "netUtils.h"

unsigned long /* ip address or 0 */
getIpAddress(const char * hostName)
{
struct hostent * host=gethostbyname(hostName); /* static alloc ! */
return host ? ntohl(*((unsigned long *)(host->h_addr))) : 0;
}

int /* nb chars written or -1 */
writeIpAddress(char * bufferOut,
               unsigned int bufferSize,
               unsigned long ipAddress)
{
int r=snprintf(bufferOut,bufferSize,"%ld.%ld.%ld.%ld",
               (ipAddress>>24)&0x00FF,
               (ipAddress>>16)&0x00FF,
               (ipAddress>>8)&0x00FF,
               ipAddress&0x00FF);
return r<0 ? -1 : r;
}

int /* >=0: file descriptor  -1: error */
openUdp(int portNumber)
{
int fd,on,r;
struct sockaddr_in addr;
socklen_t len;
RESTART_SYSCALL(fd,socket(PF_INET,SOCK_DGRAM,0));
if(fd==-1)
  {
  return -1;
  }
on=1;
RESTART_SYSCALL(r,setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(int)));
if(r==-1)
  {
  int saveErrno=errno;
  RESTART_SYSCALL(r,close(fd));
  errno=saveErrno;
  return -1;
  }
len=sizeof(struct sockaddr_in);
memset(&addr,0,len);
addr.sin_family=AF_INET;
addr.sin_port=htons(portNumber);
addr.sin_addr.s_addr=htonl(INADDR_ANY);
RESTART_SYSCALL(r,bind(fd,(const struct sockaddr *)&addr,len));
if(r==-1)
  {
  int saveErrno=errno;
  RESTART_SYSCALL(r,close(fd));
  errno=saveErrno;
  return -1;
  }
on=1;
RESTART_SYSCALL(r,setsockopt(fd,SOL_SOCKET,SO_BROADCAST,&on,sizeof(int)));
if(r==-1)
  {
  int saveErrno=errno;
  RESTART_SYSCALL(r,close(fd));
  errno=saveErrno;
  return -1;
  }
return fd;
}

int /* >=0: nb received   -1: error */
receiveUdp(int udpSocket,
           void * buffer,
           unsigned int size,
           unsigned long * ipAddressOut,
           int * portNumberOut)
{
int r;
struct sockaddr_in addr;
socklen_t len=sizeof(struct sockaddr_in);
RESTART_SYSCALL(r,recvfrom(udpSocket,buffer,size,0,
                           (struct sockaddr *)&addr,&len));
if(r!=-1)
  {
  *ipAddressOut=ntohl(addr.sin_addr.s_addr);
  *portNumberOut=ntohs(addr.sin_port);
  }
return r;
}

int /* >=0: nb sent   -1: error */
sendUdp(int udpSocket,
        const void * buffer,
        unsigned int size,
        unsigned long ipAddress,
        int portNumber)
{
int r;
struct sockaddr_in addr;
socklen_t len=sizeof(struct sockaddr_in);
memset(&addr,0,len);
addr.sin_family=AF_INET;
addr.sin_port=htons(portNumber);
addr.sin_addr.s_addr=htonl(ipAddress);
RESTART_SYSCALL(r,sendto(udpSocket,buffer,size,0,
                         (const struct sockaddr *)&addr,len));
return r;
}

int /* >=0: file descriptor  -1: error */
connectTcp(unsigned long ipAddress,
           int portNumber)
{
int fd,r;
struct sockaddr_in addr;
socklen_t len;
RESTART_SYSCALL(fd,socket(PF_INET,SOCK_STREAM,0));
if(fd==-1)
  {
  return -1;
  }
len=sizeof(struct sockaddr_in);
memset(&addr,0,len);
addr.sin_family=AF_INET;
addr.sin_port=htons(portNumber);
addr.sin_addr.s_addr=htonl(ipAddress);
RESTART_SYSCALL(r,connect(fd,(const struct sockaddr *)(&addr),len));
if(r==-1)
  {
  int saveErrno=errno;
  RESTART_SYSCALL(r,close(fd));
  errno=saveErrno;
  return -1;
  }
return fd;
}

int /* >=0: file descriptor  -1: error */
listenTcp(int portNumber)
{
int fd,on,r;
struct sockaddr_in addr;
socklen_t len;
if(portNumber<0)
  {
  portNumber=0;
  }
RESTART_SYSCALL(fd,socket(PF_INET,SOCK_STREAM,0));
if(fd==-1)
  {
  return -1;
  }
on=1;
RESTART_SYSCALL(r,setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(int)));
if(r==-1)
  {
  int saveErrno=errno;
  RESTART_SYSCALL(r,close(fd));
  errno=saveErrno;
  return -1;
  }
len=sizeof(struct sockaddr_in);
memset(&addr,0,len);
addr.sin_port=htons(portNumber);
addr.sin_addr.s_addr=htonl(INADDR_ANY);
addr.sin_family=AF_INET;
RESTART_SYSCALL(r,bind(fd,(const struct sockaddr *)&addr,len));
if(r==-1)
  {
  int saveErrno=errno;
  RESTART_SYSCALL(r,close(fd));
  errno=saveErrno;
  return -1;
  }
RESTART_SYSCALL(r,listen(fd,10));
if(r==-1)
  {
  int saveErrno=errno;
  RESTART_SYSCALL(r,close(fd));
  errno=saveErrno;
  return -1;
  }
return fd;
}

int /* >=0: file descriptor  -1: error */
acceptTcp(int listenSocket,
          unsigned long * ipAddressOut,
          int * portNumberOut)
{
struct sockaddr_in addr;
socklen_t len=sizeof(struct sockaddr_in);
int fd;
RESTART_SYSCALL(fd,accept(listenSocket,(struct sockaddr *)&addr,&len));
if(fd>=0)
  {
  *ipAddressOut=ntohl(addr.sin_addr.s_addr);
  *portNumberOut=ntohs(addr.sin_port);
  }
return fd;
}

int /* 0: success   -1: error */
getSocketAddress(int sock,
                 unsigned long * localIpAddressOut,
                 int * localPortNumberOut)
{
int r;
struct sockaddr_in addr;
socklen_t len=sizeof(struct sockaddr_in);
RESTART_SYSCALL(r,getsockname(sock,(struct sockaddr *)&addr,&len));
if(r==-1)
  {
  return -1;
  }
*localIpAddressOut=ntohl(addr.sin_addr.s_addr);
*localPortNumberOut=ntohs(addr.sin_port);
if(!*localIpAddressOut)
  {
  char hostName[0x100];
  gethostname(hostName,0x100);
  *localIpAddressOut=getIpAddress(hostName);
  }
return 0;
}

MAKE_READLINE_FUNCTION(readLine,read,int)

MAKE_READFULLY_FUNCTION(readFully,read,int)

MAKE_WRITEFULLY_FUNCTION(writeFully,write,int)

double /* current time in seconds */
getTime(void)
{
struct timeval tv;
gettimeofday(&tv,(struct timezone *)0);
return tv.tv_sec+tv.tv_usec*1e-6;
}

int /* 0: success   -1: error */
decodeUrl(const char * url,
          char * protocolOut,
          char * hostOut,
          int * portOut,
          char * fileOut)
{
int status=0;
int httpSkip=0;
if(!strncmp(url,"http:",5))
  {
  httpSkip=4;
  }
else if(!strncmp(url,"https:",6))
  {
  httpSkip=5;
  }
if(httpSkip)
  {
  const char * ptr1;
  const char * ptr2;
  const char * ptr3;
  const char * ptr4;
  const char * remaining;
  strncpy(protocolOut,url,httpSkip);
  protocolOut[httpSkip]='\0';
  ptr1=url+httpSkip+1;
  while(*ptr1=='/')
    {
    ptr1++;
    }
  ptr2=ptr1;
  while(*ptr2&&(*ptr2!=':')&&(*ptr2!='/'))
    {
    ptr2++;
    }
  strncpy(hostOut,ptr1,ptr2-ptr1);
  hostOut[ptr2-ptr1]='\0';
  *portOut=(httpSkip==4 ? 80 : 443);
  remaining=ptr2;
  if(*ptr2==':')
    {
    ptr2++;
    ptr3=ptr2;
    while(*ptr3&&(*ptr3>='0')&&(*ptr3<='9'))
      {
      ptr3++;
      }
    ptr4=ptr3;
    while(*ptr4&&(*ptr4!='/'))
      {
      ptr4++;
      }
    if((ptr3!=ptr2)&&(ptr3==ptr4))
      {
      sscanf(ptr2,"%d",portOut);
      }
    else
      {
      status=-1;
      }
    remaining=ptr4;
    }
  if(*remaining)
    {
    strcpy(fileOut,remaining);
    }
  else
    {
    strcpy(fileOut,"/");
    }
  }
else if(!strncmp(url,"file:",5))
  {
  const char * ptr;
  strcpy(protocolOut,"file");
  hostOut[0]='\0';
  *portOut=-1;
  ptr=url+5;
  if(*ptr=='/')
    {
    while(ptr[1]=='/')
      {
      ptr++;
      }
    }
  if(*ptr)
    {
    strcpy(fileOut,ptr);
    }
  else
    {
    fileOut[0]='\0';
    status=-1;
    }
  }
else /* assume url is a local file */
  {
  strcpy(protocolOut,"file");
  hostOut[0]='\0';
  *portOut=-1;
  strcpy(fileOut,url);
  }
return status;
}

int /* 0: success   -1: error */
getHttpProxy(char * proxyHostOut,
             int * proxyPortOut)
{
char * var=getenv("http_proxy");
if(!var)
  {
  var=getenv("HTTP_PROXY");
  }
if(var&&(sscanf(var,"http://%[^:]:%d",proxyHostOut,proxyPortOut)==2))
  {
  return 0;
  }
else
  {
  proxyHostOut[0]='\0';
  *proxyPortOut=-1;
  return -1;
  }
}

//----------------------------------------------------------------------------
