//----------------------------------------------------------------------------
// Copyright (c) 2007-2008 Fabrice HARROUET (ENIB)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.
// The author makes no representations about the suitability of this
// software for any purpose.
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------------

#ifndef NETUTILS_H
#define NETUTILS_H 1

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>

#define RESTART_SYSCALL(result,syscall) \
        do { (result)=(syscall); } while(((result)<0)&&(errno==EINTR));

#define MAKE_READLINE_FUNCTION(fnctName,readFnct,inputType)        \
        int /* >0: nb read   0: EOF   -1: error */                 \
        fnctName(inputType input,                                  \
                 void * buffer,                                    \
                 unsigned int size)                                \
        {                                                          \
        int r;                                                     \
        unsigned int len=0;                                        \
        char * ptr=(char *)buffer;                                 \
        --size;                                                    \
        while(len<size)                                            \
          {                                                        \
          RESTART_SYSCALL(r,readFnct(input,ptr+len,sizeof(char))); \
          if(r<0) return -1;                                       \
          if(!r||(ptr[len++]=='\n')) break;                        \
          }                                                        \
        ptr[len]='\0';                                             \
        return len;                                                \
        }

#define MAKE_READFULLY_FUNCTION(fnctName,readFnct,inputType) \
        int /* >0: nb read   0: EOF   -1: error */           \
        fnctName(inputType input,                            \
                 void * buffer,                              \
                 unsigned int size)                          \
        {                                                    \
        int r;                                               \
        unsigned int remaining=size;                         \
        char * ptr=(char *)buffer;                           \
        while(remaining)                                     \
          {                                                  \
          RESTART_SYSCALL(r,readFnct(input,ptr,remaining));  \
          if(r<0) return -1;                                 \
          if(!r) break;                                      \
          ptr+=r;                                            \
          remaining-=r;                                      \
          }                                                  \
        return size-remaining;                               \
        }

#define MAKE_WRITEFULLY_FUNCTION(fnctName,writeFnct,outputType) \
        int /* >=0: nb written   -1: error */                   \
        fnctName(outputType output,                             \
                 const void * buffer,                           \
                 unsigned int size)                             \
        {                                                       \
        int r;                                                  \
        unsigned int remaining=size;                            \
        const char * ptr=(const char *)buffer;                  \
        while(remaining)                                        \
          {                                                     \
          RESTART_SYSCALL(r,writeFnct(output,ptr,remaining));   \
          if(r<0) return -1;                                    \
          ptr+=r;                                               \
          remaining-=r;                                         \
          }                                                     \
        return size-remaining;                                  \
        }


#define FD_SET_MAX(fd,fdSet,maxFd) \
        { FD_SET((fd),(fdSet)); if((fd)>(maxFd)) { (maxFd)=(fd); } }

#define SET_TIMEVAL(tv,seconds) \
        { (tv).tv_sec=(int)(seconds); \
          (tv).tv_usec=(int)((seconds-(tv).tv_sec)*1e6); }

unsigned long /* ip address or 0 */
getIpAddress(const char * hostName);

int /* nb chars written or EOF */
writeIpAddress(char * bufferOut,
               unsigned int bufferSize,
               unsigned long ipAddress);

int /* >=0: file descriptor  -1: error */
openUdp(int portNumber);

int /* >=0: nb received   -1: error */
receiveUdp(int udpSocket,
           void * buffer,
           unsigned int size,
           unsigned long * ipAddressOut,
           int * portNumberOut);

int /* >=0: nb sent   -1: error */
sendUdp(int udpSocket,
        const void * buffer,
        unsigned int size,
        unsigned long ipAddress,
        int portNumber);

int /* >=0: file descriptor  -1: error */
connectTcp(unsigned long ipAddress,
           int portNumber);

int /* >=0: file descriptor  -1: error */
listenTcp(int portNumber);

int /* >=0: file descriptor  -1: error */
acceptTcp(int listenSocket,
          unsigned long * ipAddressOut,
          int * portNumberOut);

int /* 0: success   -1: error */
getSocketAddress(int sock,
                 unsigned long * localIpAddressOut,
                 int * localPortNumberOut);

int /* >0: nb read   0: EOF   -1: error */
readLine(int fd,
         void * buffer, 
         unsigned int size);

int /* >0: nb read   0: EOF   -1: error */
readFully(int fd,
          void * buffer,
          unsigned int size);

int /* >=0: nb written   -1: error */
writeFully(int fd,
           const void * buffer,
           unsigned int size);

double /* current time in seconds */
getTime(void);

int /* 0: success   -1: error */
decodeUrl(const char * url,
          char * protocolOut,
          char * hostOut,
          int * portOut,
          char * fileOut);

int /* 0: success   -1: error */
getHttpProxy(char * proxyHostOut,
             int * proxyPortOut);

#endif /* NETUTILS_H */

//----------------------------------------------------------------------------
